/*************************************************************
 * File         : user-src/src/runktau/runktau.c
 * Version      : $Id: runktauTrace.c,v 1.1 2006/11/12 01:24:52 anataraj Exp $
 *
 ************************************************************/

/************************************************
 * KTAU Performance Analysis Suite		*
 * 						*
 * runktau:					*
 * 	- Launches any program,			*
 * 	- waits for termination,		*
 * 	- reads Kernel Profile/Trace,		*
 * 	- and dumps to file.			*
 *						*
 * 	- Functionality similar to 'time' cmd	*
 * 						*
 * Usage: runktau <full program path> [args] [args]
 * 						*
 ***********************************************/

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/utsname.h>

#include <fcntl.h>

#include "ktau_proc_interface.h"
#include "ktau_proc_map.h"
#include "ktau_and_tau.h"
#include "ktau_diff_profiles.h"

//#include "../ktaud/ktau_timer.c"

/* Status Field */
#define S_INPROG	0x0
#define S_DONE		0x1

#define POLL_INTV	2400	/*Poll every X secs for result */
//#define POLL_INTV	1	

#define FNAME_SIZE 256

volatile int status = S_INPROG;

volatile int numchildren = 0;

double tsc = 0;

char* g_sym_map_file;
int g_getaggr = 0;
char *g_start_buffer = NULL;
int g_start_size;
int g_istrace = 0;

#ifdef BGL_CNLINUX
//substitute routines
#define _printf my_printf
#define _perror my_perror

//instead of inclusing "my.h", we declare the below...
void init_zoid(int argc, char** argv, char** envp);
ssize_t my_writev (int fd, const struct iovec *vector, int count);
ssize_t my_write(int fd, const void *buf, size_t count);
ssize_t my_read(int fd, void *buf, size_t count);
ssize_t my_readv (int fd, const struct iovec *vector, int count);
int my_open(const char *pathname, ...);
int my_close(int fd);
void my_printf(const char *format, ...);
void myfprintf(FILE *stream, const char *format, ...);
void my_perror(const char *string);
void init_zoid(int argc, char** argv, char** envp);
void mount_devices();

//from "unix.h"
int unix_open(const char *pathname /* in:str */,
              int flags /* in:obj */,
              mode_t mode /* in:obj */);
ssize_t unix_write(int fd /* in:obj */,
                   const void *buf /* in:arr:size=+1:zerocopy */,
                   size_t count /* in:obj */);

int g_trace_f = 0;
typedef int g_trace_type;

#else /* BGL_CNLINUX */

//substitute routines
#define _printf printf
#define _perror perror

FILE* g_trace_f = NULL;
typedef FILE* g_trace_type;
#endif

void dump_ktau_data(int pid, int type, char* map_file, int getaggr, char* start_buffer, int start_sz); 
void dump_ktau_trace(int pid, int getaggr, g_trace_type fp);

double ktau_get_tsc(void);

static void _sigchld(int readpid)
{
	int pid;
	int saved_errno = errno;

	//_printf("Calling waitpid with: %d\n", readpid);

	while ((pid = waitpid(readpid, 0, WNOHANG)) != 0) {

		//_printf("Waitpid returned: %d\n",pid);

		if (pid == -1) {
			if (errno != EINTR)
				break;
		} else
			++numchildren;
	}

	errno = saved_errno;
	status = S_DONE;
}

static void sigchld() {
	_sigchld(-1);
}

static void sigchld_sigaction(int signo, siginfo_t *psiginfo, void* parg) {
	int pid = 0;

	if(signo != SIGCHLD) {
		_printf("Non-SigChld Signal: %d. Ignoring.\n", signo);
		return;
	}

	//_printf("SigChld Recieved....\n");

	if(!psiginfo) {
		_printf("psigninfo is NULL. Ignoring.\n");
		return;
	}

	pid = psiginfo->si_pid;

	_printf("Pid form sigingfo: %d\n", pid);

	if(!g_istrace) {
		dump_ktau_data(pid, KTAU_PROFILE, g_sym_map_file, g_getaggr, g_start_buffer, g_start_size);
	} else {
		dump_ktau_trace(pid, g_getaggr, g_trace_f);
	}

	_sigchld(pid);
}


void usage(char* cmd) {
	fprintf(stderr, "USAGE:\n");
	fprintf(stderr, "prompt>%s <System.map path> <System-wide Data collection Yes(1) or No(0)> <TRACE or PROFILE> <No. Entries to Resize Trace Buffer - 0 means dont> <executable path> <arg 1> <arg 2> <...>\n\n", cmd);
	fprintf(stderr, "EXAMPLE 1: Profile /usr/bin/find\n");
	fprintf(stderr, "prompt>%s /boot/System.map 0 PROFILE 0 /usr/bin/find . -name \"ktau\"\n\n", cmd);
	fprintf(stderr, "EXAMPLE 2: Profile /usr/bin/find AND REST of system\n");
	fprintf(stderr, "prompt>%s /boot/System.map 1 PROFILE 0 /usr/bin/find . -name \"ktau\"\n\n", cmd);
	fprintf(stderr, "EXAMPLE 3: TRACE /usr/bin/find AND REST of system\n");
	fprintf(stderr, "prompt>%s /boot/System.map 1 TRACE 0 /usr/bin/find . -name \"ktau\"\n\n", cmd);
	fprintf(stderr, "EXAMPLE 4: TRACE /usr/bin/find AND REST of system AND Enlarge Trace buffer size of the 'find' process to 409600 entries.\n");
	fprintf(stderr, "prompt>%s /boot/System.map 1 TRACE 40960 /usr/bin/find . -name \"ktau\"\n\n", cmd);
	fflush(stderr);
}


int main(int argc, char* argv[])
{
	struct sigaction act;
	int pid;
	char *parent="Parent:", *child="Child:";
	char* name = parent, *sym_map_file = NULL;
	int getaggr = 0;
	int istrace = 0, trace_resize = 0;
	char trace_fname[1024];

#ifdef BGL_CNLINUX
	mount_devices();
	_printf("Initializing zoid communication\n");
	init_zoid(argc, argv, 0);
	my_printf("Hello, World! (from zoid-compatible _printf! inside cn-linux->runktauTrace!)\n\n");
	_printf("Hello, World! (from zoid-compatible _printf! inside cn-linux->runktauTrace!)\n\n");
#endif /* BGL_CNLINUX */

	if( (argc < 6) || ((argc > 1) && (!strcmp(argv[1],"-h")))) {
		usage(argv[0]); 
		exit(-1);
	}

	sym_map_file = argv[1];
	g_sym_map_file = argv[1];

	getaggr = atoi(argv[2]);
	g_getaggr = getaggr;

	if(!strcmp(argv[3], "TRACE")) {
		istrace = 1;
	} else if(!strcmp(argv[3], "PROFILE")) {
		istrace = 0;
	} else {
		_printf("Bad argument: %s - should be TRACE or PROFILE\n.", argv[3]);
		exit(-1);
	}

	g_istrace = istrace;

	trace_resize = atoi(argv[4]);

	tsc = ktau_get_tsc();
	act.sa_sigaction = sigchld_sigaction;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_SIGINFO;
	sigaction(SIGCHLD, &act, 0);

	if(!g_istrace) {
		if(getaggr) {
			g_start_size = get_ktau_profs(0, KTAU_PROFILE, getaggr, &g_start_buffer); 
		}
	}

	switch((pid=fork())) {
		case -1:
			_perror("fork");
			return 1;
		case 0:
			name = child;
			sleep(5);
			_printf("%s] Error: Execv returned: %d\n", name, execv(argv[5], argv+5));
			_exit(0);
			break;
		default:
			if(g_istrace) {
				#ifdef BGL_CNLINUX
				sprintf(trace_fname,"/home/anataraj/OUTPUT/CNLINUX-DETOUR/Ktrace.%d",pid);
				g_trace_f = unix_open(trace_fname, O_CREAT | O_RDWR, 0777);
				#else
				sprintf(trace_fname,"Ktrace.%d",pid);
				g_trace_f = fopen(trace_fname, "w");
				#endif
				if(!g_trace_f) {
					_perror("Trace File fopen:");
					exit(-1);
				}

				dump_ktau_trace(pid, g_getaggr, g_trace_f);

				//resize to what was requested
				if(trace_resize) {
					_printf(" resize ret: %d\n", ktau_trace_resize(trace_resize, KTAU_TRACE, 0, &pid, 1, 0, NULL));
				}
			}
			_printf("%s Parent: Child Pid is: %d\n", name, pid);
			while(status == S_INPROG) {
				sleep(POLL_INTV);
				if(g_istrace) {
					dump_ktau_trace(pid, g_getaggr, g_trace_f);
				}
			}
			break;
	}

	//_printf("%s Number of children = %d\n", name, numchildren);
	return 0;
}


int get_ktau_profs(int pid, int type, int getaggr, char** pbuffer) {
	char *buffer;
        ktau_output* output = NULL;
        int size = 0, ret_val = 0, rd_size = 0, noprofs = 0;

	size = read_size(type, 0, &pid, !(getaggr), 0, NULL, -1.0);

	if(size <= 0) {
		_printf("get_ktau_profs: read_size ret bad size: %d.\n", size);
		return -1;
	}

	buffer = (char*)malloc(size);
	if(!buffer) {
		_printf("get_ktau_profs: malloc failed. \n");
		return -1;
	}


        /* Read Data*/
        ret_val = read_data(type, 0, &pid, !(getaggr), buffer, size,
                        0 /*tolerance*/,
                        NULL /*flags*/);
        if(ret_val < 0) {
                _perror("get_ktau_profs: read_data failure.");
		goto free_out;
        }

	*pbuffer = buffer;

	return ret_val;

free_out:
	free(buffer);
	buffer = NULL;
	*pbuffer = NULL;
        return 0;
}

void dump_ktau_trace(int pid, int getaggr, g_trace_type dump_fs) {
	char *buffer, fname[FNAME_SIZE];
        ktau_output* output = NULL, *start_output = NULL;
        int size = 0, ret_val = 0, rd_size = 0, noprofs = 0, start_noprofs = 0, nodiff_profs = 0;

	rd_size = get_ktau_profs(pid, KTAU_TRACE, getaggr, &buffer);

	if(rd_size <= 0) {
		_printf("dump_ktau_trace: get_ktau_profs ret bad size: %d.\n", rd_size);
		return;
	}

#ifdef BGL_CNLINUX
	ret_val = size = 0;
	while(rd_size > 0) {
		ret_val = unix_write(dump_fs, (char*)(buffer + size), rd_size);
		if(ret_val < 0) {
			_printf("unix_write returns -ve: %d\n", ret_val);
			_perror("dump_ktau_trace->unix_write");
			exit(-1);
		}
		rd_size -= ret_val;
		size += ret_val;
	}
#else /* BGL_CNLINUX */
        noprofs = unpack_bindata(KTAU_TRACE, buffer, rd_size, &output);
        if(!noprofs || !output) {
                _printf("ktau_dump_trace: stop ktau_output from unpack_bindata is:%x. no-profs is:%d\n", output, noprofs);
		goto free_out;
        }

	s_printf(fname, "trace.%d",pid);

	if(!dump_fs) {
		perror("dump_ktau_trace: fopen:");
		return;
	}

	print_many_profiles(KTAU_TRACE, dump_fs, output, noprofs);
#endif /* BGL_CNLINUX */

	free(buffer);

	return;

free_out:
	free(buffer);
	_printf("dump_ktau_trace: Error. Aborting.\n");
	exit(-1);
}

void dump_ktau_data(int pid, int type, char* sym_map_file, int getaggr, char* start_buffer, int start_sz) {
	char *buffer, fname[FNAME_SIZE];
        ktau_output* output = NULL, *start_output = NULL;
        int size = 0, ret_val = 0, rd_size = 0, noprofs = 0, start_noprofs = 0, nodiff_profs = 0;
	FILE* dump_fs = NULL;

	struct ktau_sym_map* psymmap = NULL;

	ktau_output* paggr_start_prof = NULL;
	ktau_output* paggr_prof = NULL;
	ktau_output* paggr_diff = NULL;
	ktau_output* pdiff_profs = NULL;

	rd_size = get_ktau_profs(pid, type, getaggr, &buffer);

	if(rd_size <= 0) {
		_printf("dump_ktau_data: get_ktau_profs ret bad size: %d.\n", rd_size);
		return;
	}

	if(getaggr) {
		/* Unpack start Data*/
		start_noprofs = unpack_bindata(type, start_buffer, start_sz, &start_output);
		if(!start_noprofs || !start_output) {
			_printf("ktau_dump_data: start ktau_output from unpack_bindata is:%x. no-profs is:%d\n", start_output, start_noprofs);
			goto free_out;
		}
	}
        noprofs = unpack_bindata(type, buffer, rd_size, &output);
        if(!noprofs || !output) {
                _printf("ktau_dump_data: stop ktau_output from unpack_bindata is:%x. no-profs is:%d\n", output, noprofs);
		goto free_out;
        }
	
	/*
	s_printf(fname, "./Kprofile.%d", pid);
	dump_fs = fopen(fname, "w");
	if(!dump_fs) {
		_printf("dump_ktau_data: Cannot open %s. fopen returned Null.", fname);
		goto free_out;
	}
       
	fprintf(dump_fs,"TSC: %f\n",tsc);
	fprintf(dump_fs,"No Profiles: 1\n");
	print_ktau_output(type, dump_fs, output,0);
	fprintf(dump_fs,"TSC: %f\n",tsc);
	fprintf(dump_fs,"No Profiles: 1\n");
	print_ktau_output(type, dump_fs, output);
	*/
	
	psymmap = ktau_get_kallsyms(sym_map_file);
	if(!psymmap) {
		_printf("ktau_get_kallsyms ret error.\n");
		goto free_out;
	}


	if(getaggr) {
		paggr_start_prof = (ktau_output*)calloc(sizeof(ktau_output),1);
		if(!paggr_start_prof) {
			_printf("calloc ret null.\n");
			goto free_out;
		}
		paggr_start_prof->ent_lst = (o_ent*)calloc(sizeof(o_ent)*2048,1);
		if(!paggr_start_prof->ent_lst) {
			_printf("calloc ret null.\n");
			free(paggr_start_prof);
			goto free_out;
		}
		aggr_many_profiles(start_output, start_noprofs, 2048, paggr_start_prof);
		paggr_start_prof->pid = -1;

		paggr_prof = (ktau_output*)calloc(sizeof(ktau_output),1);
		if(!paggr_prof) {
			_printf("calloc ret null.\n");
			goto free_out;
		}
		paggr_prof->ent_lst = (o_ent*)calloc(sizeof(o_ent)*2048,1);
		if(!paggr_prof->ent_lst) {
			_printf("calloc ret null.\n");
			free(paggr_prof);
			goto free_out;
		}
		aggr_many_profiles(output, noprofs, 2048, paggr_prof);
		paggr_prof->pid = -1;

		ktau_diff_profiles(paggr_start_prof, 1, paggr_prof, 1, &paggr_diff);
		print_tau_profiles(paggr_diff, 1, -1, psymmap, tsc/1000000); //we want usec resolution, not secs
	}

	if(!getaggr) {
		print_tau_profiles(output, noprofs, pid, psymmap, tsc/1000000); //we want usec resolution, not secs
	} else {
		nodiff_profs = ktau_diff_profiles(start_output, start_noprofs, output, noprofs, &pdiff_profs);
		print_tau_profiles(pdiff_profs, nodiff_profs, pid, psymmap, tsc/1000000); //we want usec resolution, not secs
	}

free_out:
	free(buffer);
	buffer = NULL;

        return;
}

